﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;
using LibUsbDotNet;
using LibUsbDotNet.Main;
using LibUsbDotNet.DeviceNotify;
using Salamax.Logging;

namespace Salamax
{
  public class BridgeService
  {
    private WiiRemote _wiiRemote = new WiiRemote();
    private SilverlinkCable _silverlink = new SilverlinkCable();
    private Xbox360Controller _controller = new Xbox360Controller();
    private TeensySimulatorDevice _teensy = new TeensySimulatorDevice();

    public BridgeService()
    {
      _silverlink.ReceivedStaticDataRequest += _silverlink_ReceivedStaticDataRequest;
      _silverlink.ReceivedChallenge1 += _silverlink_ReceivedChallenge1;
      _silverlink.ReceivedChallenge2 += _silverlink_ReceivedChallenge2;
      ButtonStateChanges.ButtonStateChanged += ButtonStateChanges_ButtonStateChanged;

      ButtonStateChanges.AddNotifier(_wiiRemote);
    }

    public WiiRemote Wiimote
    {
      get
      {
        return _wiiRemote;
      }
    }

    public Xbox360Controller Controller
    {
      get
      {
        return _controller;
      }
    }

    public bool IsSilverlinkConnected
    {
      get
      {
        return _silverlink.IsConnected;
      }
    }

    public bool IsTeensyConnected
    {
      get
      {
        return _teensy.IsConnected;
      }
    }

    public bool IsControllerConnected
    {
      get
      {
        return _controller.IsConnected;
      }
    }

    public void Start()
    {
      Start(null, null);
    }

    public void Start(int? vendorId, int? productId)
    {
      try { _wiiRemote.Connect(); }
      catch { /* Eat it, because the Wiimote library is a little dumb... */ };

      //Detect all the necessary devices
      _FindAllDevices(vendorId, productId);

      Logger.WriteLine("Bridge service started.", Logger.Type.Information);
    }

    public void ForwardButtonStateChanges()
    {
      ButtonStateChanges.RefreshState();
    }

    public void Stop()
    {
      //Disconnect all devices
      _wiiRemote.Disconnect();
      _controller.Disconnect();
      _silverlink.Disconnect();
      _teensy.Disconnect();

      Logger.WriteLine("Bridge service stopped.", Logger.Type.Information);
    }

    private void _FindAllDevices(int? vendorId, int? productId)
    {
      if (vendorId.HasValue && productId.HasValue)
        _controller.Find(vendorId.Value, productId.Value);
      else
        _controller.Find();
      _silverlink.Find();
      _teensy.Find();
    }

    private void _silverlink_ReceivedChallenge1(object sender, ChallengeEventArgs e)
    {
      var buffer = _controller.DoChallenge1(e.Data);
      _silverlink.ForwardData(0x83, buffer, buffer.Length);
    }

    private void _silverlink_ReceivedChallenge2(object sender, ChallengeEventArgs e)
    {
      var buffer = _controller.DoChallenge2(e.Data);
      _silverlink.ForwardData(0x83, buffer, buffer.Length);
    }

    private void _silverlink_ReceivedStaticDataRequest(object sender, EventArgs e)
    {
      if (_controller != null)
      {
        //Get the data from the controller
        var buffer = _controller.RequestStaticData();

        //Also get all the descriptors from the controller
        var device = new byte[0x12];
        int transferred1, transferred2, transferred3;
        var setup = new UsbSetupPacket(0x80, 0x06, 0x0100, 0x0000, 0x12);
        _controller.DoControlTransfer(ref setup, ref device, device.Length, out transferred1);
        setup = new UsbSetupPacket(0x80, 0x06, 0x0200, 0x0000, 0x1000);
        var config = new byte[0x1000];
        _controller.DoControlTransfer(ref setup, ref config, config.Length, out transferred2);
        //Find the right string descriptor index
        int i = 0;
        int strIndex = 0;
        while (i < config.Length)
        {
          if (config[i + 1] == 0x04)
          {
            if (config[i + 5] == 0xFF && config[i + 6] == 0xFD && config[i + 7] == 0x13)
            {
              strIndex = config[i + 8];
              break;
            }
          }

          i += Math.Max((int)config[i + 0], 1);
        }
        var str = new byte[0x1000];
        setup = new UsbSetupPacket(0x80, 0x06, (short)((short)0x0300 | (short)strIndex), 0x0409, 0x1000);
        _controller.DoControlTransfer(ref setup, ref str, str.Length, out transferred3);
        var all = new byte[1 + buffer.Length + transferred1 + transferred2 + transferred3];
        all[0] = (byte)(buffer.Length & 0xFF);
        Array.Copy(buffer, 0, all, 1, buffer.Length);
        Array.Copy(device, 0, all, 1 + buffer.Length, transferred1);
        Array.Copy(config, 0, all, 1 + buffer.Length + transferred1, transferred2);
        Array.Copy(str, 0, all, 1 + buffer.Length + transferred1 + transferred2, transferred3);

        //Send the response back
        _silverlink.ForwardData(0x81, all, all.Length);
      }
    }

    private void ButtonStateChanges_ButtonStateChanged(object sender, EventArgs e)
    {
      if (_silverlink != null && _silverlink.IsConnected)
        _silverlink.ForwardData(0x00, ButtonStateChanges.KeyStateInfo, ButtonStateChanges.KeyStateInfo.Length);

      if (_teensy != null && _teensy.IsConnected)
        _teensy.SendKeyData(ButtonStateChanges.KeyStateInfo);
    }
  }
}
