﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading;
using System.Windows.Forms;
using Salamax.Logging;

namespace Salamax
{
  public static class Startup
  {
    [DllImport("user32.dll")]
    public static extern IntPtr FindWindow(string lpClassName, string lpWindowName);

    [DllImport("user32.dll")]
    static extern bool ShowWindow(IntPtr hWnd, int nCmdShow);

    [DllImport("user32.dll")]
    static extern IntPtr SetFocus(IntPtr hWnd);

    public static BridgeService Service { get; private set; }
    private static string _scriptFileName = String.Empty;
    private static bool _scriptStarted = false;

    static void Main(string[] args)
    {
      try
      {
        /* Code to trigger LibUsbDotNet bug/quirk/whatever.
        const int _VENDOR_ID = 0x0E6F;
        const int _PRODUCT_ID = 0x0213;
        var d = LibUsbDotNet.UsbDevice.OpenUsbDevice(new LibUsbDotNet.Main.UsbDeviceFinder(_VENDOR_ID, _PRODUCT_ID)) as LibUsbDotNet.IUsbDevice;
        var writer = d.OpenEndpointWriter(LibUsbDotNet.Main.WriteEndpointID.Ep02);
        int transferred;
        writer.Write(new byte[] { 0x01, 0x03, 0x00 }, 0, 3, 5000, out transferred);
        writer.Flush();
        writer.Dispose();
        writer = null;
        d.Close();
        d = null;
        return;
         */

        /* USB simulator test code.
        var device = new USBSimulator.USBSimulatorDevice();
        var controller = new SimulatedController();
        device.CurrentDevice = controller;
        device.AttachDevice();
        while (true)
        {
          if (Console.KeyAvailable)
          {
            var key = Console.ReadKey(true).Key;

            if (key == ConsoleKey.Escape)
              break;
            if (key == ConsoleKey.RightArrow)
              controller.KeyData[2] |= 0x08;
          }

          System.Threading.Thread.Sleep(10);
          controller.KeyData[2] = 0x00;
        }
        controller.Stop();
        device.DetachDevice();
        System.Threading.Thread.Sleep(3000);
        device.CurrentDevice = null;
        controller.OnShutdown();
        device.Close();
        LibUsbDotNet.LibUsb.LibUsbDevice.Exit();
        return;
         */

        //Process any command-line arguments
        int? vendorId = null;
        int? productId = null;
        int raw;
        foreach (var arg in args)
        {
          if (arg.StartsWith("/script="))
          {
            _scriptFileName = arg.Replace("/script=", String.Empty);
            if (!File.Exists(_scriptFileName))
              throw new ArgumentException("Specified script file does not exist");
          }
          else if (arg.StartsWith("/vid="))
          {
            if (int.TryParse(arg.Replace("/vid=", String.Empty), System.Globalization.NumberStyles.HexNumber, null, out raw))
              vendorId = raw;
          }
          else if (arg.StartsWith("/pid="))
          {
            if (int.TryParse(arg.Replace("/pid=", String.Empty), System.Globalization.NumberStyles.HexNumber, null, out raw))
              productId = raw;
          }
        }

        if (!String.IsNullOrEmpty(_scriptFileName))
          _Run(false, vendorId, productId);
        else
          _Run(true, vendorId, productId);

        LibUsbDotNet.LibUsb.LibUsbDevice.Exit();
      }
      catch (Exception ex)
      {
        Console.WriteLine("ERROR: " + ex.ToString());
      }
    }

    private static void _Run(bool gui)
    {
      _Run(gui, null, null);
    }

    private static void _Run(bool gui, int? vendorId, int? productId)
    {
      try
      {
        const string TITLE = "Salamax Console Window";
        Form frm = null;

        if (gui)
        {
          //Set up this window
          Console.Title = TITLE;

          //Show the main form
          frm = new Main();
          frm.Show();
          frm.Activate();

          Logger.WriteLine("Running in GUI mode.", Logger.Type.Information);
        }

        //Start the bridge service!
        Service = new BridgeService();
        Service.Start(vendorId, productId);

        //Display instructions if necessary
        var commands = new Queue<string>();
        var notifier = new CodeNotifier();
        if (!String.IsNullOrEmpty(_scriptFileName))
        {
          foreach (var line in File.ReadAllLines(_scriptFileName))
            if (!String.IsNullOrEmpty(line))
              commands.Enqueue(line);

          ButtonStateChanges.AddNotifier(notifier);

          Console.WriteLine("Press [ENTER] to begin executing script.");
        }

        bool done = false;
        while (!done)
        {
          //Let any UI catch up
          try
          {
            Application.DoEvents();
          }
          catch
          {
          }

          if (gui)
          {
            //Hide the console window if it is somehow visible...
            IntPtr ptr = FindWindow(null, TITLE);
            if (ptr != IntPtr.Zero)
              ShowWindow(ptr, 0);

            //If our main form is gone, then so are we
            if (!Utilities.IsFormOpen(frm))
              done = true;
          }
          else
          {
            if (Console.KeyAvailable)
            {
              var key = Console.ReadKey().Key;

              if (key == ConsoleKey.Escape)
                done = true;
              else if (!String.IsNullOrEmpty(_scriptFileName) && key == ConsoleKey.Enter)
                _scriptStarted = true;
            }
          }

          if (_scriptStarted)
          {
            if (commands.Count > 0)
            {
              var cmd = commands.Dequeue();
              var parameters = new List<string>();
              if (cmd.IndexOf(" ") != -1)
                foreach (var p in cmd.Substring(cmd.IndexOf(" ") + 1).Split(new char[] { ' ' }))
                  parameters.Add(p);

              //Deal with the standard stuff (one or more parameters)
              if (parameters.Count > 0)
              {
                byte byte1; byte.TryParse(parameters[0], out byte1);
                short short1; short.TryParse(parameters[0], out short1);

                if (cmd.StartsWith("Sleep"))
                  Thread.Sleep(Convert.ToInt32(parameters[0]));
                else if (cmd.StartsWith("LeftTrigger"))
                  notifier.LeftTrigger = byte1;
                else if (cmd.StartsWith("RightTrigger"))
                  notifier.RightTrigger = byte1;
                else if (cmd.StartsWith("LeftX"))
                  notifier.LeftX = short1;
                else if (cmd.StartsWith("LeftY"))
                  notifier.LeftY = short1;
                else if (cmd.StartsWith("RightX"))
                  notifier.RightX = short1;
                else if (cmd.StartsWith("RightY"))
                  notifier.RightY = short1;

                bool? held = null;
                if (cmd.StartsWith("ButtonDown")) held = true;
                if (cmd.StartsWith("ButtonUp")) held = false;

                if (held.HasValue)
                {
                  switch (parameters[0])
                  {
                    case "A":
                      notifier.A = held.Value;
                      break;
                    case "B":
                      notifier.B = held.Value;
                      break;
                    case "X":
                      notifier.X = held.Value;
                      break;
                    case "Y":
                      notifier.Y = held.Value;
                      break;
                    case "Start":
                      notifier.Start = held.Value;
                      break;
                    case "Back":
                      notifier.Back = held.Value;
                      break;
                    case "Guide":
                      notifier.Guide = held.Value;
                      break;
                    case "Up":
                      notifier.Up = held.Value;
                      break;
                    case "Down":
                      notifier.Down = held.Value;
                      break;
                    case "Left":
                      notifier.Left = held.Value;
                      break;
                    case "Right":
                      notifier.Right = held.Value;
                      break;
                    case "LeftButton":
                      notifier.LeftButton = held.Value;
                      break;
                    case "RightButton":
                      notifier.RightButton = held.Value;
                      break;
                    default:
                      break;
                  }
                }
              }
            }
            else
            {
              Console.WriteLine("Script has finished executing.");

              _scriptStarted = false;
              done = true;
            }
          }

          Service.ForwardButtonStateChanges();

          Thread.Sleep(10);
        }

        Service.Stop();
      }
      catch (Exception ex)
      {
        Logger.WriteLine("Error while running bridge service: " + ex.ToString(), Logger.Type.Error);
      }
    }
  }
}
